@@ -1,9 +1,39 @@
 #!/usr/bin/env python
 
-import gtk
 import sys
 import os
 import glob
+import gtk
+import tl.eggdeps.graph
+
+
+class ExtraDiscerningGraph(tl.eggdeps.graph.Graph):
+    """Dependency graph that remembers extras.
+
+    tl.eggdeps.graph.Graph discards the extras components from requirement
+    specifications, so if package A requires package X, and package B requires
+    package X with some extras, the graph nodes for A and B will be identical.
+
+    (It remembers information going in the other direction, e.g. if package X
+    requires package M, but if you want package X with an extra, it will also
+    require package N, then graph['X'] will be dict(M=set(), N=set('extra'))).
+
+    This doesn't suit me, so I override names() to create fake nodes of the
+    form 'project_name[extra]'.  As a result, if a package B requres package X
+    with [extra1, extra2], you'll see three names in the specs list:
+
+        graph['B'] = {'X': set(), 'X[extra1]': set(), 'X[extra2]': set()}
+
+    """
+
+    def names(self, specs):
+        names = set()
+        for s in specs:
+            if self.show(s.project_name):
+                names.add(s.project_name)
+                for extra in s.extras:
+                    names.add(s.project_name + '[' + extra + ']')
+        return names
 
 
 class MyTreeModel(gtk.GenericTreeModel):
@@ -14,22 +44,46 @@ class MyTreeModel(gtk.GenericTreeModel):
 
     def __init__(self, graph):
         gtk.GenericTreeModel.__init__(self)
-        self.graph = graph
-        self.children = {None: sorted(graph.roots)}
-        for node_id, node in self.graph.iteritems():
-            self.children[node_id] = sorted(node)
+        self.build_graph(graph)
         self.calc_depths()
         self.calc_dep_sizes()
 
+    def build_graph(self, graph):
+        self.children = {None: sorted(graph.roots)}
+        self.node_name = {}
+        self.node_available = {}
+        for node_id, node in graph.iteritems():
+            node_name = node.name
+            if not node.compatible:
+                node_name = '(%s)' % node_name
+            self.add_node(node_id, node_name, node.compatible)
+            if node.dist:
+                for extra in sorted(node.dist.extras):
+                    self.add_node(node_id + '[' + extra +']',
+                                  node_name + ' [' + extra + ']',
+                                  node.compatible)
+            for child_id, extras in sorted(node.items()):
+                if extras:
+                    for extra in sorted(extras):
+                        self.add_edge(node_id + '[' + extra + ']', child_id)
+                else:
+                    self.add_edge(node_id, child_id)
+
+    def add_node(self, node_id, node_name, available):
+        assert node_id not in self.children
+        self.children[node_id] = []
+        self.node_name[node_id] = node_name
+        self.node_available[node_id] = available
+
+    def add_edge(self, node_id, child_id):
+        self.children[node_id].append(child_id)
+
     def calc_depths(self):
-        graph = self.graph
-        self.depth = {}
-        for root in graph.roots:
-            self.depth[root] = 1
-        queue = list(graph.roots)
+        self.depth = {None: 0}
+        queue = [None]
         while queue:
             u = queue.pop(0)
-            for v in graph[u]:
+            for v in self.children[u]:
                 if v not in self.depth:
                     self.depth[v] = self.depth[u] + 1
                     queue.append(v)
@@ -37,12 +91,12 @@ class MyTreeModel(gtk.GenericTreeModel):
     def calc_dep_sizes(self):
         # Use Brute Force When In Doubt
         self.dep_sizes = {}
-        for o in self.graph:
+        for o in self.children:
             reachable = set()
             stack = [o]
             while stack:
                 u = stack.pop()
-                for v in self.graph[u]:
+                for v in self.children[u]:
                     if v not in reachable:
                         reachable.add(v)
                         stack.append(v)
@@ -71,12 +125,6 @@ class MyTreeModel(gtk.GenericTreeModel):
             node_id = self.children[node_id][n]
         return node_id
 
-    def _node(self, rowref):
-        node_id = None
-        for n in rowref:
-            node_id = self.children[node_id][n]
-        return self.graph[node_id]
-
     def _node_with_extra_info(self, rowref):
         recursive = False
         seen = set()
@@ -88,22 +136,18 @@ class MyTreeModel(gtk.GenericTreeModel):
             else:
                 seen.add(node_id)
         excessive_depth = len(rowref) > self.depth[node_id]
-        return self.graph[node_id], recursive, excessive_depth
+        return node_id, recursive, excessive_depth
 
     def on_get_value(self, rowref, column):
         if column == 0:
-            node = self._node(rowref)
-            if node.compatible:
-                return node.name
-            else:
-                return '(%s)' % node.name
+            return self.node_name[self._node_id(rowref)]
         elif column == 1:
-            node, recursive, excessive = self._node_with_extra_info(rowref)
+            node_id, recursive, excessive = self._node_with_extra_info(rowref)
             if recursive:
                 return 'red'
             elif excessive:
                 return 'blue'
-            elif not node.compatible:
+            elif not self.node_available[node_id]:
                 return 'gray'
         elif column == 2:
             return self.dep_sizes[self._node_id(rowref)]
@@ -116,13 +160,13 @@ class MyTreeModel(gtk.GenericTreeModel):
         return rowref[:-1] + (idx, )
 
     def on_iter_children(self, parent):
-        if not self._node(parent):
+        if not self.children[self._node_id(parent)]:
             return None # no children
         return parent + (0, )
 
     def on_iter_has_child(self, rowref):
-        node, recursive, excessive_depth = self._node_with_extra_info(rowref)
-        return bool(node) and not recursive
+        node_id, recursive, excessive_depth = self._node_with_extra_info(rowref)
+        return bool(self.children[node_id]) and not recursive
 
     def on_iter_n_children(self, rowref):
         return len(self.children[self._node_id(rowref)])
@@ -172,8 +216,7 @@ class MainWindow(object):
         return self.model._node_with_extra_info(path)[2]
 
     def load_tree(self, specs=()):
-        import tl.eggdeps.graph
-        graph = tl.eggdeps.graph.Graph(show=lambda name: name != 'setuptools')
+        graph = ExtraDiscerningGraph(show=lambda name: name != 'setuptools')
         if specs:
             self.window.set_title("Egg dependencies for %s" % " ".join(specs))
             graph.from_specifications(*specs)
